/*
 * include/asm-i386/i387.h
 *
 * Copyright (c) 1999 Doug Ledford <dledford@redhat.com>
 *
 * Made from various code bits pulled from other files
 * in order to put things together in a way that made
 * sense.
 *
 * FX/FPU support:
 * Copyright (c) 1999 Ingo Molnar <mingo@redhat.com>,
 *                   Gabriel Paubert <paubert@iram.es>
 */
#include <linux/config.h>

#ifndef __ASM_I386_I387_H
#define __ASM_I386_I387_H

extern int i387_hard_to_user ( struct user_i387_struct * user,
	union i387_hard_union * hard);
extern int i387_user_to_hard ( union i387_hard_union * hard,
	struct user_i387_struct * user);

/*
 * Fill out the reserved bits, treat it as an fsave struct since the
 * union makes this work for both fsave and fxsave structs.
 */
#ifdef CONFIG_X86_FX

#define i387_save_hard(x) \
do { \
	if (boot_cpu_data.x86_capability & X86_FEATURE_FXSR) { \
		__asm__ __volatile__("fxsave %0" \
				     : "=m" ((x).hard.fxsave.fxcwd)); \
	} else { \
		__asm__ __volatile__("fnsave %0; fwait;" \
				     : "=m" ((x).hard.fsave.cwd)); \
	} \
} while(0)

#define i387_restore_hard(x) \
do { \
	if (boot_cpu_data.x86_capability & X86_FEATURE_FXSR) { \
                __asm__ __volatile__("fxrstor %0" \
				     : \
				     : "m" ((x).hard.fxsave.fxcwd)); \
	} else { \
		__asm__ __volatile__("frstor %0" \
				     : \
				     :"m" ((x).hard.fsave.cwd)); \
	} \
} while(0)

#define i387_set_cwd(x,v) \
do { \
	if (boot_cpu_data.x86_capability & X86_FEATURE_FXSR) { \
		(x).fxsave.fxcwd = (short)(v); \
	} else { \
		(x).fsave.cwd = ((long)(v) | 0xffff0000); \
	} \
} while(0)

#define i387_get_swd(x,v) \
do { \
	if (boot_cpu_data.x86_capability & X86_FEATURE_FXSR) { \
		(v) = (unsigned long) (x).fxsave.fxswd; \
	} else { \
		(v) = (unsigned long) (x).fsave.swd; \
	} \
} while(0)

#define i387_set_swd(x,v) \
do { \
	if (boot_cpu_data.x86_capability & X86_FEATURE_FXSR) { \
		(x).fxsave.fxswd = (short)(v); \
	} else { \
		(x).fsave.swd = ((long)(v) | 0xffff0000); \
	} \
} while(0)

#define i387_set_twd(x,v) \
do { \
	if (boot_cpu_data.x86_capability & X86_FEATURE_FXSR) { \
		(x).fxsave.fxtwd = (short)(v); \
	} else { \
		(x).fsave.twd = ((long)(v) | 0xffff0000); \
	} \
} while(0)

static inline unsigned short fputag_KNI_to_387(unsigned char tb) {
	unsigned short tw = tb;
	tw = (tw | (tw << 4)) & 0x0f0f; /* zzzz7654zzzz3210 */
	tw = (tw | (tw << 2)) & 0x3333; /* zz76zz54zz32zz10 */
	tw = (tw | (tw << 1)) & 0x5555; /* z7z6z5z4z3z2z1z0 */
	tw = ~(tw * 3);
	return tw;
}

static inline unsigned char fputag_387_to_KNI(unsigned short tw) {
	tw = ~tw & 0x5555;		/* z7z6z5z4z3z2z1z0 */
	tw = (tw | (tw >> 1)) & 0x3333;	/* zz76zz54zz32zz10 */
	tw = (tw | (tw >> 2)) & 0x0f0f;	/* zzzz7654zzzz3210 */
	tw = (tw | (tw >> 4)) & 0x00ff; /* zzzzzzzz76543210 */
	return tw;
}

#else /* CONFIG_X86_FX */

#define i387_save_hard(x) \
do { \
	__asm__ __volatile__("fnsave %0; fwait;" \
			     : "=m" ((x).hard.fsave.cwd)); \
} while(0)

#define i387_restore_hard(x) \
do { \
	__asm__ __volatile__("frstor %0" \
			     : \
			     :"m" ((x).hard.fsave.cwd)); \
} while(0)

#define i387_set_cwd(x,v) \
do { (x).fsave.cwd = ((long)(v) | 0xffff0000); } while(0)

#define i387_get_swd(x,v) \
do { (v) = (unsigned long) (x).fsave.swd; } while(0)

#define i387_set_swd(x,v) \
do { (x).fsave.swd = ((long)(v) | 0xffff0000); } while(0)

#define i387_set_twd(x,v) \
do { (x).fsave.twd = ((long)(v) | 0xffff0000); } while(0)

#endif /* CONFIG_X86_FX */


/*
 * For when we want to use the FPU in kernel code
 * 
 * These functions allow the use of up to 4 KNI based xmm registers on the
 * Pentium III processors or up to 4 MMX registers on Pentium MMX and above
 * or compatible processors.  Pick the routines that you need based on the
 * regs you are going to use.  Keep in mind that these are intended to be
 * used only after you've verified that the processor supports these
 * operations.  Use them before you've done that and watch your machine go
 * boom.  Take a look in arch/i386/lib/best_function.c for an example of
 * how to fixup the kernel with kni/mmx using functions once the CPU
 * capabilities have been determined.
 *
 * In all of these functions:
 *
 *   recursive - int, used to determine what the state is at restore time
 *   regs - char * to an array that is 32 bytes for mmx and 64 bytes for kni
 *          which is then used to save off the contents of the current
 *          regs to be recursively safe
 *   task_switch_regs - char * to another array of the same size as the one
 *          above, but this array is optional.  If your function might get 
 *          pre-empted by another task then this pointer should be non-NULL
 *          so that at unlazy_fpu() time in the switch_to() function we
 *          can save your register state (copy_*_user functions are an example
 *          of functions that need this, since they can take a page fault and
 *          while that fault is being serviced the scheduler is free to run
 *          another task entirely).
 *   irqflags - unsigned long used to store IRQ state
 */

#define SAVE_MMX_REGS(num_regs, regs) \
	switch( (num_regs) ) { \
		case 1: \
			__asm__ __volatile__("movq %%mm0, 0x00(%0)\n\t" \
					     : : "r" ((regs)) : "memory" ); \
			break; \
		case 2: \
			__asm__ __volatile__("movq %%mm0, 0x00(%0)\n\t" \
					     "movq %%mm1, 0x08(%0)\n\t" \
					     : : "r" ((regs)) : "memory" ); \
			break; \
		case 3: \
			__asm__ __volatile__("movq %%mm0, 0x00(%0)\n\t" \
					     "movq %%mm1, 0x08(%0)\n\t" \
					     "movq %%mm2, 0x10(%0)\n\t" \
					     : : "r" ((regs)) : "memory" ); \
			break; \
		case 4: \
			__asm__ __volatile__("movq %%mm0, 0x00(%0)\n\t" \
					     "movq %%mm1, 0x08(%0)\n\t" \
					     "movq %%mm2, 0x10(%0)\n\t" \
					     "movq %%mm3, 0x18(%0)\n\t" \
					     : : "r" ((regs)) : "memory" ); \
			break; \
		default: \
			panic("Invalid number of regs passed to SAVE_MMX_REGS\n"); \
	}

#define RESTORE_MMX_REGS(num_regs, regs) \
	switch( (num_regs) ) { \
		case 1: \
			__asm__ __volatile__("movq 0x00(%0), %%mm0\n\t" \
					     : : "r" ((regs)) : "memory" ); \
			break; \
		case 2: \
			__asm__ __volatile__("movq 0x00(%0), %%mm0\n\t" \
					     "movq 0x08(%0), %%mm1\n\t" \
					     : : "r" ((regs)) : "memory" ); \
			break; \
		case 3: \
			__asm__ __volatile__("movq 0x00(%0), %%mm0\n\t" \
					     "movq 0x08(%0), %%mm1\n\t" \
					     "movq 0x10(%0), %%mm2\n\t" \
					     : : "r" ((regs)) : "memory" ); \
			break; \
		case 4: \
			__asm__ __volatile__("movq 0x00(%0), %%mm0\n\t" \
					     "movq 0x08(%0), %%mm1\n\t" \
					     "movq 0x10(%0), %%mm2\n\t" \
					     "movq 0x18(%0), %%mm3\n\t" \
					     : : "r" ((regs)) : "memory" ); \
			break; \
		default: \
			panic("Invalid number of regs passed to RESTORE_MMX_REGS\n"); \
	}

#define SAVE_KNI_REGS(num_regs, regs) \
	switch( (num_regs) ) { \
		case 1: \
			__asm__ __volatile__("movups %%xmm0, 0x00(%0)\n\t" \
					     : : "r" ((regs)) : "memory" ); \
			break; \
		case 2: \
			__asm__ __volatile__("movups %%xmm0, 0x00(%0)\n\t" \
					     "movups %%xmm1, 0x10(%0)\n\t" \
					     : : "r" ((regs)) : "memory" ); \
			break; \
		case 3: \
			__asm__ __volatile__("movups %%xmm0, 0x00(%0)\n\t" \
					     "movups %%xmm1, 0x10(%0)\n\t" \
					     "movups %%xmm2, 0x20(%0)\n\t" \
					     : : "r" ((regs)) : "memory" ); \
			break; \
		case 4: \
			__asm__ __volatile__("movups %%xmm0, 0x00(%0)\n\t" \
					     "movups %%xmm1, 0x10(%0)\n\t" \
					     "movups %%xmm2, 0x20(%0)\n\t" \
					     "movups %%xmm3, 0x30(%0)\n\t" \
					     : : "r" ((regs)) : "memory" ); \
			break; \
		default: \
			panic("Invalid number of regs passed to SAVE_KNI_REGS\n"); \
	}

#define RESTORE_KNI_REGS(num_regs, regs) \
	switch( (num_regs) ) { \
		case 1: \
			__asm__ __volatile__("movups 0x00(%0), %%xmm0\n\t" \
					     : : "r" ((regs)) : "memory" ); \
			break; \
		case 2: \
			__asm__ __volatile__("movups 0x00(%0), %%xmm0\n\t" \
					     "movups 0x10(%0), %%xmm1\n\t" \
					     : : "r" ((regs)) : "memory" ); \
			break; \
		case 3: \
			__asm__ __volatile__("movups 0x00(%0), %%xmm0\n\t" \
					     "movups 0x10(%0), %%xmm1\n\t" \
					     "movups 0x20(%0), %%xmm2\n\t" \
					     : : "r" ((regs)) : "memory" ); \
			break; \
		case 4: \
			__asm__ __volatile__("movups 0x00(%0), %%xmm0\n\t" \
					     "movups 0x10(%0), %%xmm1\n\t" \
					     "movups 0x20(%0), %%xmm2\n\t" \
					     "movups 0x30(%0), %%xmm3\n\t" \
					     : : "r" ((regs)) : "memory" ); \
			break; \
		default: \
			panic("Invalid number of regs passed to RESTORE_KNI_REGS\n"); \
	}

#define SFENCE() \
	__asm__ __volatile__("sfence":::"memory")


extern spinlock_t kern_fpu_lock;

/*
 *  BIG FAT WARNING ---- DO NOT MIX KNI AND MMX OPTIMIZATIONS IN THE SAME
 *  KERNEL OR ELSE USER SPACE FPU REGS WILL SUFFER AS A RESULT!!!!!!!!!!!
 *
 *  SECOND BIG FAT WARNING ---- EVEN THOUGH THE num_regs VARIABLE ONLY NEEDS
 *  TO BE AS LARGE AS THE REGS YOU ARE USING, THE task_switch_regs ARRAY
 *  MUST BE LARGE ENOUGH TO HOLD 4 OF THE REGISTERS YOU ARE USING BECAUSE
 *  unlazy_fpu HAS NO IDEA HOW MANY REGS WE ARE USING.  SINCE 4 IS THE
 *  MAXIMUM NUMBER OF REGS ALLOWED, WE MUST HAVE SPACE FOR 4 ON A TASK
 *  SWITCH.
 */

/*
 * FPU lazy state save handling..
 */
#define clear_fpu(tsk) do { \
	unsigned long irqflags; \
	spin_lock_irqsave(&kern_fpu_lock, irqflags); \
	tsk->used_math=0; \
	tsk->tss.x86_fpustate=0; \
	stts(); \
	spin_unlock_irqrestore(&kern_fpu_lock, irqflags); \
} while (0)

#define save_kern_fpu(tsk) do { \
	if(tsk->tss.mmx_reg_space != NULL) \
		__asm__("movq %%mm0, 0x00(%0)\n\t" \
			"movq %%mm1, 0x08(%0)\n\t" \
			"movq %%mm2, 0x10(%0)\n\t" \
			"movq %%mm3, 0x18(%0)\n\t" \
			:: "r" (tsk->tss.mmx_reg_space):"memory"); \
	if(tsk->tss.xmm_reg_space != NULL) \
		__asm__("movups %%xmm0, 0x00(%0)\n\t" \
			"movups %%xmm1, 0x10(%0)\n\t" \
			"movups %%xmm2, 0x20(%0)\n\t" \
			"movups %%xmm3, 0x30(%0)\n\t" \
			:: "r" (tsk->tss.xmm_reg_space):"memory"); \
} while (0)

#define unlazy_fpu(tsk) do { \
	unsigned long irqflags; \
	spin_lock_irqsave(&kern_fpu_lock, irqflags); \
	if (tsk->tss.x86_fpustate & X86_FPUSTATE_FPU_ENABLED) { \
		if (tsk->tss.x86_fpustate & X86_FPUSTATE_KERN_ANY) { \
			save_kern_fpu(tsk); \
		} else if (!(tsk->tss.x86_fpustate & X86_FPUSTATE_FPU_SAVED) && \
			    (tsk->used_math)) { \
			i387_save_hard(tsk->tss.i387); \
			tsk->tss.x86_fpustate |= X86_FPUSTATE_FPU_SAVED; \
		} \
		tsk->tss.x86_fpustate &= ~X86_FPUSTATE_FPU_ENABLED; \
		stts(); \
	} \
	spin_unlock_irqrestore(&kern_fpu_lock, irqflags); \
} while (0)

#define save_user_fpu_context(tsk) do { \
	if (!((tsk)->tss.x86_fpustate & X86_FPUSTATE_FPU_ENABLED)) { \
		clts(); \
		(tsk)->tss.x86_fpustate |= X86_FPUSTATE_FPU_ENABLED; \
		if((tsk)->tss.mmx_reg_space) { \
			__asm__("movq 0x00(%0), %%mm0\n\t" \
				"movq 0x08(%0), %%mm1\n\t" \
				"movq 0x10(%0), %%mm2\n\t" \
				"movq 0x18(%0), %%mm3\n\t" \
				:: "r" ((tsk)->tss.mmx_reg_space)); \
		} else if((tsk)->tss.xmm_reg_space) { \
			__asm__("movups 0x00(%0), %%xmm0\n\t" \
				"movups 0x10(%0), %%xmm1\n\t" \
				"movups 0x20(%0), %%xmm2\n\t" \
				"movups 0x30(%0), %%xmm3\n\t" \
				:: "r" ((tsk)->tss.xmm_reg_space)); \
		} \
	} else if (!((tsk)->tss.x86_fpustate & X86_FPUSTATE_FPU_SAVED) && \
		    ((tsk)->used_math)) { \
		i387_save_hard((tsk)->tss.i387); \
		(tsk)->tss.x86_fpustate |= X86_FPUSTATE_FPU_SAVED; \
	} \
} while (0)
	
#define kernel_take_fpu_mmx(recursive, num_regs, regs, task_switch_regs) do { \
	unsigned long irqflags; \
	struct task_struct *tsk=current; \
	spin_lock_irqsave(&kern_fpu_lock, irqflags); \
	(recursive) = tsk->tss.x86_fpustate & X86_FPUSTATE_KERN_ANY; \
	save_user_fpu_context(tsk); \
	if((recursive) & X86_FPUSTATE_KERN_MMX) { \
		SAVE_MMX_REGS((num_regs), (regs)); \
	} else { \
		tsk->tss.x86_fpustate |= X86_FPUSTATE_KERN_MMX; \
	} \
	if ((task_switch_regs) && !tsk->tss.mmx_reg_space) { \
		tsk->tss.mmx_reg_space = (task_switch_regs); \
	} \
	spin_unlock_irqrestore(&kern_fpu_lock, irqflags); \
} while (0)

#define kernel_release_fpu_mmx(recursive, num_regs, regs, task_switch_regs) do { \
	unsigned long irqflags; \
	struct task_struct *tsk=current; \
	spin_lock_irqsave(&kern_fpu_lock, irqflags); \
	if (tsk->tss.mmx_reg_space == task_switch_regs) { \
		tsk->tss.mmx_reg_space = NULL; \
	} \
	if((recursive) & X86_FPUSTATE_KERN_MMX) { \
		if (!(tsk->tss.x86_fpustate & X86_FPUSTATE_FPU_ENABLED)) { \
			clts(); \
			tsk->tss.x86_fpustate |= X86_FPUSTATE_FPU_ENABLED; \
		} \
		RESTORE_MMX_REGS((num_regs), (regs)); \
	} else { \
		tsk->tss.x86_fpustate &= ~X86_FPUSTATE_KERN_MMX; \
		if(!(recursive) && \
		    (tsk->tss.x86_fpustate & X86_FPUSTATE_FPU_ENABLED)) { \
			tsk->tss.x86_fpustate &= ~X86_FPUSTATE_FPU_ENABLED; \
			stts(); \
		} \
	} \
	spin_unlock_irqrestore(&kern_fpu_lock, irqflags); \
} while (0)

#define kernel_take_fpu_kni(recursive, num_regs, regs, task_switch_regs) do { \
	unsigned long irqflags; \
	struct task_struct *tsk=current; \
	spin_lock_irqsave(&kern_fpu_lock, irqflags); \
	(recursive) = tsk->tss.x86_fpustate & X86_FPUSTATE_KERN_ANY; \
	save_user_fpu_context(tsk); \
	if((recursive) & X86_FPUSTATE_KERN_KNI) { \
		SAVE_KNI_REGS((num_regs), (regs)); \
	} else { \
		tsk->tss.x86_fpustate |= X86_FPUSTATE_KERN_KNI; \
	} \
	if ((task_switch_regs) && !tsk->tss.xmm_reg_space) { \
		tsk->tss.xmm_reg_space = (task_switch_regs); \
	} \
	spin_unlock_irqrestore(&kern_fpu_lock, irqflags); \
} while (0)
		
	
#define kernel_release_fpu_kni(recursive, num_regs, regs, task_switch_regs) do { \
	unsigned long irqflags; \
	struct task_struct *tsk=current; \
	spin_lock_irqsave(&kern_fpu_lock, irqflags); \
	if (tsk->tss.xmm_reg_space == task_switch_regs) { \
		tsk->tss.xmm_reg_space = NULL; \
	} \
	if((recursive) & X86_FPUSTATE_KERN_KNI) { \
		if(!(tsk->tss.x86_fpustate & X86_FPUSTATE_FPU_ENABLED)) { \
			clts(); \
			tsk->tss.x86_fpustate |= X86_FPUSTATE_FPU_ENABLED; \
		} \
		RESTORE_KNI_REGS((num_regs), (regs)); \
	} else { \
		tsk->tss.x86_fpustate &= ~X86_FPUSTATE_KERN_KNI; \
		if(!(recursive) && \
		    (tsk->tss.x86_fpustate & X86_FPUSTATE_FPU_ENABLED)) { \
			tsk->tss.x86_fpustate &= ~X86_FPUSTATE_FPU_ENABLED; \
			stts(); \
		} \
	} \
	spin_unlock_irqrestore(&kern_fpu_lock, irqflags); \
} while (0)


#endif /* __ASM_I386_I387_H */
