#!/usr/bin/perl -w
#
#  $Id: versioncheck.pl,v 1.1 2000/04/25 15:38:37 jeffb Exp $
# 
# Copyright 2000 Cobalt Networks, Inc http://www.cobalt.com/

use strict;

# wrap this around the packing_list, and output new packing
# list as we verify rpm versioning to resolve conflicts.
# accounts for UPDATE[1-9] also.

# hardcoded stuff
my $pack = 'packing_list';     # location of packing list
my $nupk = 'packing_list.upd'; # temp packing list we'll create
my $rpms = 'RPMS';             # location of RPMS
my $base = shift @ARGV;        # base directory for pkg install
$ENV{HOME} = '/root';          # for the sanity of rpm

# Main

# process root
chdir ( $base ) or die "Failed to change to directory $base: $!\n";
stripit( $pack, $nupk, "." );

# look for update directories
for( 1..9 ) {
  if ( -e "UPDATE$_" ) {
    stripit( $pack,  $nupk, "UPDATE$_" );
  }
}


# stripit()
#
# overview: tests and removes rpms for packing_list and directory
# inputs:   packing list, temp packing list, and base directory
# returns:  0 on success 1 on error
sub stripit {
  my( $oldpl, $newpl, $basdir ) = @_;

  open( OFH, "$basdir/$oldpl" )    or die "Failed to open $oldpl: $!\n";
  open( NFH, ">$basdir/$newpl" )   or die "Failed to open $newpl: $!\n";

  OUT: while( <OFH> ) {
    chomp;

    if ( /^RPM/ ) {
      /RPM: (\S+)/;  # RPM name is stored in $1
      my $rpm = $1;
      my $return = `rpm -U $basdir/$rpms/$1 --test 2>&1`;

      CASE: {
        if ( $return =~ /already/ )  { # This rpm is already installed
                                       deleteit("$basdir/$rpms/$rpm");
                                       next OUT;
                                     };
        if ( $return =~ /newer/ )    { # A newer version of this rpm is installed
                                       deleteit("$basdir/$rpms/$rpm");
                                       next OUT;
                                     };
       }

    }
    print NFH $_, "\n";
  }

  close( OFH );
  close( NFH );

  rename( "$basdir/$newpl", "$basdir/$oldpl" );

  return;
}

# deleteit()
#
# overview: deletes specified file
# inputs:   EXACTLY one filename, no lists
# returns:  0 on success 1 on error
sub deleteit($) {
  my $filename = shift;

  if ( ! -e $filename ) { return 1; }  # oops, no file there

  unlink ( $filename );

  return;
}
